module Completion where

import Data.List
import Term
import Rule
import TRS
import Order
import Control.Monad

data Env = Env { 
  _kTRSs :: Int, 
  _nEquations :: Int, 
  _sizeBound :: Int,
  _tool :: String,
  _order :: Order.Class
}

-- trsOrSC env E C returns either a complete TRS for E or S(C).
data CompleteTRS_or_Ext = 
    CompleteTRS (TRS, Order.Instance) 
  | Ext TRS deriving Show

type Record = [(Rule, (Int, Int))]

subsumeEq :: Rule -> Rule -> Bool
subsumeEq (s, t) (u, v) =
  match' [] [(s, u), (t, v)] /= Nothing ||
  match' [] [(s, v), (t, u)] /= Nothing

encompassEq :: Rule -> Rule -> Bool
encompassEq e1 e2
  | subsumeEq e1 e2 = True
encompassEq e1 (F h us, F k vs) = 
  h == k && any (encompassEq e1) (zip us vs)
encompassEq _ _ = False

encompassEq' :: Rule -> [Term] -> [Term] -> Bool
encompassEq' e1 (u : us) (v : vs)
  | u == v    = encompassEq' e1 us vs
  | otherwise = encompassEq e1 (u, v) && us == vs
encompassEq' _ _ _ = False

minimal :: TRS -> TRS
minimal []       = []
minimal (e : es) = e : minimal [ e' | e' <- es, not (encompassEq e e') ]

type History = [(Rule, (Int,Int))]

-- Completion without the deduce rule.

nfEq :: TRS -> Rule -> Rule
nfEq trs (s, t) = (nf trs s, nf trs t)

interreduce :: Rule -> TRS -> (TRS, TRS)
interreduce _ [] = ([], [])
interreduce rule ((l, r) : trs)
  | Just l' <- rewrite [rule] l = ((l', r) : cs, trs')
  | otherwise                   = (cs, (l, r) : trs')
  where (cs, trs') = interreduce rule trs

orient :: Order.Instance -> Rule -> Maybe Rule
orient oi (s, t)
  | Order.gt oi s t = Just (s, t)
  | Order.gt oi t s = Just (t, s)
  | otherwise       = Nothing

simplify :: Order.Instance -> TRS -> TRS -> TRS -> (TRS, TRS)
simplify _  []       unoriented trs = (unoriented, trs)
simplify oi (e : cs) unoriented trs
  | s == t    = simplify oi cs unoriented trs
  | otherwise =
      case orient oi e' of
        Nothing   -> simplify oi cs (e' : unoriented) trs
        Just rule -> simplify oi (cs' ++ cs ++ unoriented) [] (rule : trs')
          where (cs', trs') = interreduce rule trs
  where
    e'@(s, t) = nfEq trs e

psi :: Order.Instance -> TRS -> (TRS, TRS)
psi oi cs = simplify oi cs [] []

score :: History -> Rule -> (Int, Int)
score h rule
  | Just (_, i) <- lookup rule h = (Rule.size rule, i)
  | otherwise                    = (Rule.size rule, length h)

select :: Env -> History -> TRS -> TRS
select (Env { _nEquations = n }) h cs =
  take n (sortOn (score h) (minimal [ e | e <- cs ]))

completeTRS_or_Ext' :: Env -> History -> TRS -> [(TRS,Order.Instance)] -> TRS -> IO CompleteTRS_or_Ext
completeTRS_or_Ext' env h cs [] ext =
  return (Ext (select env h (diffES ext cs)))
completeTRS_or_Ext' env h cs ((_, oi) : pairs) ext
  | cs0 == [] && licps' == [] = return (CompleteTRS (trs', oi))
  | otherwise = completeTRS_or_Ext' env h cs pairs ext'
  where
    (cs0, trs') = psi oi cs
    licps = licp trs'
    licps' = normalizeES trs' licps
    cs' = nubES (cs0 ++ trs' ++ licps')
    ext' = unionES ext cs'
    
completeTRS_or_Ext :: Env -> History -> TRS -> [(TRS,Order.Instance)] -> IO CompleteTRS_or_Ext
completeTRS_or_Ext env h cs pairs = completeTRS_or_Ext' env h cs pairs []

-- The main partial function of maximal completion.

elemEq :: Rule -> TRS -> Bool
elemEq (s, t) trs = elem (s, t) trs || elem (t, s) trs

lookupEq :: Rule -> History -> String
lookupEq _    [] = "## ##: "
lookupEq rule ((rule', (j, i)) : a)
  | Rule.variantEq rule rule' = show j ++ " " ++ show i ++ ": "
  | otherwise                 = lookupEq rule a

printC :: History -> TRS -> IO ()
printC h cs = forM_ cs f
  where f e = putStrLn (lookupEq e h ++ "  " ++ showEq e)

phi :: Env -> History -> Order.Class -> Measure -> Int -> TRS -> TRS -> IO (Maybe (TRS, Order.Instance))
phi env h oc m i es cs = do
  pairs <- kOrders (_tool env) oc m (_kTRSs env) cs
  case pairs of
    [] -> return Nothing
    _ -> do
      trsOrExt <- completeTRS_or_Ext env h cs pairs
      case trsOrExt of
        CompleteTRS pair -> do
          return (Just pair)
        Ext ext -> do
          if subES ext cs then do
            if length pairs == _kTRSs env then do
              phi (env { _kTRSs = _kTRSs env + 1 }) h oc m (i + 1) es cs
            else
              return Nothing
          else do
            let cs' = diffES ext cs
            let n = length h
            let h' = h ++ [ (e, (n + j, i)) | (j, e) <- zip [1..] cs' ]
            let cs'' = unionES cs ext
            phi env h' oc m (i + 1) es (minimal cs'')

complete :: Env -> TRS -> IO (Maybe (TRS, Order.Instance))
complete env es = do
  m <- phi env h (_order env) Redundancy 1 es es
  case m of
    Nothing  -> return Nothing
    Just pair -> return (Just pair)
  where 
    h = [ (e, (j, 0)) | (j, e) <- zip [1..] es ]
