module EKBO where

import Data.List (elemIndex)
import Term
import TRS
import Rule
import Algebra
import SMT hiding (variables)

type Weight = ([(String, Int)], Int)

type Parameter = (Algebra, Weight, Precedence)

showWeight :: Weight -> String
showWeight (wf, w0) =
  unlines (( "  w0 = " ++ show w0 ) :
           [ "  w(" ++ f ++ ") = " ++ show k | (f, k) <- wf ])

showParameter :: Parameter -> String
showParameter (i, w, p) =
  "EKBO with interpretations on natural numbers\n\n" ++
  showInterpretations i ++
  "\nweights\n\n" ++
  showWeight w ++
  "\nand precedence:\n\n" ++
  showPrecedence p ++ "\n"

-- w(t)
weight :: Weight -> Term -> Int
weight (_, w0) (V _)    = w0
weight w@(wf, _) (F f ts)
  | Just n <- lookup f wf = n + sum [ weight w t | t <- ts ]
  | otherwise             = error "weight"

validWeight1 :: Weight -> (String, Int) -> Bool
validWeight1 (wf, w0) (f, n)
  | Just k <- lookup f wf = k >= 0 && (n /= 0 || k >= w0) 
  | otherwise             = False

validWeight :: Weight -> Signature -> Bool
validWeight w@(_, w0) sig = w0 > 0 && all (validWeight1 w) sig

largest :: Precedence -> String -> Bool
largest (g : _) f = f == g
largest _       _ = False

admissible1 :: Parameter -> (String, Int) -> Bool
admissible1 (_, (wf, _), p) (f, n)
  | Just k <- lookup f wf = n /= 1 || k > 0 || largest p f
  | otherwise = error "admissible1"

admissible :: Parameter -> Signature -> Bool
admissible param sig = all (admissible1 param) sig

gtWeight :: Weight -> Term -> Term -> Bool
gtWeight w s t = weight w s > weight w t

-- exponentialForm t f x checks if t is of form f^n(x).
exponentialForm :: Term -> String -> String -> Bool
exponentialForm (V y)     _ x = x == y
exponentialForm (F g [t]) f x = f == g && exponentialForm t f x
exponentialForm _ _ _         = False

gtKBO :: Parameter -> Term -> Term -> Bool
gtKBO param@(_, w, _) s t = 
  nonduplicating (s, t) && 
  ws > wt || (ws >= wt && gtKBO' param s t) 
  where
    ws = weight w s
    wt = weight w t

gtKBO' :: Parameter -> Term -> Term -> Bool
gtKBO' _ t@(F f _) (V x) = exponentialForm t f x
gtKBO' param@(a, _, p) s@(F _ ss) t@(F _ ts) =
  gtB p a s t ||
  (geqB a s t && gtLex (gtKBO param) ss ts)
gtKBO' _ _ _ = False

gt :: Parameter -> Term -> Term -> Bool
gt param@(a, _, _) s t = geqA a s t && gtKBO param s t

varW0 :: Exp
varW0 = Var "w0"

varWf :: String -> Exp
varWf f = Var ("wf" ++ f)

encodeWeight :: Term -> Exp
encodeWeight (V _)    = varW0
encodeWeight (F f ts) = Plus (varWf f : [ encodeWeight t | t <- ts ])

encodeGtWeight :: Term -> Term -> Formula
encodeGtWeight s t = Gt (encodeWeight s) (encodeWeight t)

encodeGeqWeight :: Term -> Term -> Formula
encodeGeqWeight s t = Geq (encodeWeight s) (encodeWeight t)

encodeGtLex :: (Term -> Term -> Formula) -> [Term] -> [Term] -> Formula
encodeGtLex _ [] [] = bottom
encodeGtLex encodeGt (s : ss) (t : ts)
  | s == t    = encodeGtLex encodeGt ss ts
  | otherwise = encodeGt s t
encodeGtLex _ _ _ = error "encodeGtLex"

encodeGtKBO :: Term -> Term -> Formula
encodeGtKBO s t
  | nonduplicating (s, t) = 
      disj [encodeGtWeight s t, 
            conj [encodeGeqWeight s t, encodeGtKBO' s t]]
  | otherwise = bottom

encodeGtKBO' :: Term -> Term -> Formula
encodeGtKBO' s@(F f _) (V x)
  | exponentialForm s f x = top
encodeGtKBO' s@(F _ ss) t@(F _ ts) =
  disj [encodeGtB s t,
        conj [encodeGeqB s t, encodeGtLex encodeGtKBO ss ts]]
encodeGtKBO' _ _ = bottom

encodeGtEKBO :: Term -> Term -> Formula
encodeGtEKBO s t = conj [encodeGeqA s t, encodeGtKBO s t] 

encodeValidityOfWeight :: Signature -> Formula
encodeValidityOfWeight sig =
  conj (Gt varW0 (Val 0) :
        [ Geq (varWf f) varW0   | (f, 0) <- sig ] ++
        [ Geq (varWf f) (Val 0) | (f, n) <- sig, n > 0 ])

encodeAdmissibility1 :: Signature -> String -> Formula
encodeAdmissibility1 sig f =
  disj [Gt (varWf f) (Val 0),
        conj (neg (encodeLabeled f) : [ encodeGtPrec f g | (g, _) <- sig, f /= g ]) ]

encodeAdmissibility :: Signature -> Formula
encodeAdmissibility sig = 
  conj [ encodeAdmissibility1 sig f | (f, 1) <- sig ] 

varGtEKBO :: TRS -> Term -> Term -> Formula
varGtEKBO a l r
  | Just i <- elemIndex (l, r) a = FVar ("gtEKBO" ++ show i)
  | otherwise = error "varGtEKBO"

encode :: Signature -> TRS -> (Formula, [(Rule, Formula)])
encode sig trs =
  (conj (Distinct [ varPrec f | (f, _) <- sig ] :
         encodeValidityOfWeight sig :
         encodeAdmissibility sig :
         [ Geq (varA f) (Val 0) | (f, _) <- sig ] ++
         [ Geq (varA (mark f)) (Val 0) | (f, _) <- sig ] ++
         [ iff (varGtEKBO trs l r) (encodeGtEKBO l r) | (l, r) <- trs ] ),
   [ ((l, r), varGtEKBO trs l r) | (l, r) <- trs ])


-- Proving termination by SSPO 

decodeWeight :: Signature -> Model -> Weight
decodeWeight sig model =
  ([ (f, evalExp model (varWf f)) | (f, _) <- sig ],
   evalExp model varW0)

decode :: Signature -> Model -> Parameter
decode sig model =
  (decodeAlgebra model sig, decodeWeight sig model, decodePrecedence model)
