module KBO where

import Data.List (elemIndex)
import Data.Maybe (fromMaybe)
import Term
import Rule
import TRS
import Algebra
import SMT hiding (variables)

type Weight = ([(String, Int)], Int)

type Parameter = (Weight, Precedence)

showWeight :: Weight -> String
showWeight (wf, w0) =
  unlines (( "  w0 = " ++ show w0 ) :
           [ "  w(" ++ f ++ ") = " ++ show k | (f, k) <- wf ])

showParameter :: Parameter -> String
showParameter (w, p) =
  "KBO with weight\n\n" ++
  showWeight w ++
  "\nand precedence:\n\n" ++
  showPrecedence p ++ "\n"

-- exponentialForm t f x checks if t is of form f^n(x).
exponentialForm :: Term -> String -> String -> Bool
exponentialForm (V y)     _ x = x == y
exponentialForm (F g [t]) f x = f == g && exponentialForm t f x
exponentialForm _ _ _         = False

-- w(t)
weight :: Weight -> Term -> Int
weight (_, w0) (V _)    = w0
weight w@(wf, _) (F f ts)
  | Just n <- lookup f wf = n + sum [ weight w t | t <- ts ]
  | otherwise             = error "weight"

gt :: Parameter -> Term -> Term -> Bool
gt param@(w, _) s t = 
  nonduplicating (s, t) && 
  ws > wt || (ws >= wt && gt' param s t) 
  where
    ws = weight w s
    wt = weight w t

gt' :: Parameter -> Term -> Term -> Bool
gt' _ t@(F f _) (V x) = exponentialForm t f x
gt' param@(_, p) (F f ss) (F g ts) =
  gtPrec p f g ||
  (f == g && gtLex (gt param) ss ts)
gt' _ _ _ = False

-- Encoding functions for SSPO based on KBO.

idW0 :: String
idW0 = "w0"

varW0 :: Exp
varW0 = Var idW0

idWf :: String -> String
idWf f = "wf" ++ f

varWf :: String -> Exp
varWf f = Var (idWf f)

encodeWeight :: Term -> Exp
encodeWeight (V _)    = varW0
encodeWeight (F f ts) = Plus (varWf f : [ encodeWeight t | t <- ts ])

encodeGtWeight :: Term -> Term -> Formula
encodeGtWeight s t = Gt (encodeWeight s) (encodeWeight t)

encodeGeqWeight :: Term -> Term -> Formula
encodeGeqWeight s t = Geq (encodeWeight s) (encodeWeight t)

encodeGtLex :: (Term -> Term -> Formula) -> [Term] -> [Term] -> Formula
encodeGtLex _ [] [] = bottom
encodeGtLex encodeGt (s : ss) (t : ts)
  | s == t    = encodeGtLex encodeGt ss ts
  | otherwise = encodeGt s t
encodeGtLex _ _ _ = error "encodeGtLex"

encodeGtKBO :: Term -> Term -> Formula
encodeGtKBO s t
  | nonduplicating (s, t) = 
      disj [encodeGtWeight s t, 
            conj [encodeGeqWeight s t, encodeGtKBO' s t]]
  | otherwise = bottom

encodeGtKBO' :: Term -> Term -> Formula
encodeGtKBO' s@(F f _) (V x)
  | exponentialForm s f x = top
encodeGtKBO' (F f ss) (F g ts)
  | f == g    = encodeGtLex encodeGtKBO ss ts
  | otherwise = encodeGtPrec f g
encodeGtKBO' _ _ = bottom

encodeValidityOfWeight :: Signature -> Formula
encodeValidityOfWeight sig =
  conj (Gt varW0 (Val 0) :
        [ Geq (varWf f) varW0   | (f, 0) <- sig ] ++
        [ Geq (varWf f) (Val 0) | (f, n) <- sig, n > 0 ])

encodeAdmissibility1 :: Signature -> String -> Formula
encodeAdmissibility1 sig f =
  disj [Gt (varWf f) (Val 0),
        conj [ encodeGtPrec f g | (g, _) <- sig, f /= g ] ]

encodeAdmissibility :: Signature -> Formula
encodeAdmissibility sig = 
  conj [ encodeAdmissibility1 sig f | (f, 1) <- sig ] 

idGtKBO :: TRS -> Term -> Term -> String
idGtKBO a l r
  | Just i <- elemIndex (l, r) a = "rule" ++ show i
  | otherwise = error "idKBO"

varGtKBO :: TRS -> Term -> Term -> Formula
varGtKBO a l r = FVar (idGtKBO a l r)

encode :: Signature -> TRS -> (Formula, [(Rule, Formula)])
encode sig trs =
  (conj (Distinct [ varPrec f | (f, _) <- sig ] :
         encodeValidityOfWeight sig :
         encodeAdmissibility sig :
         [ iff (varGtKBO trs l r) (encodeGtKBO l r) | (l, r) <- trs ] ),
   [ ((s, t), varGtKBO trs s t) | (s, t) <- trs ])

-- Proving termination by SSPO 

decodeWeight :: Model -> Signature -> Weight
decodeWeight model sig =
  ([ (f, fromMaybe (wf n) (lookup (idWf f) model)) | (f, n) <- sig ], w0)
  where 
    w0 = fromMaybe 1 (lookup idW0 model)
    wf 0 = w0
    wf 1 = w0
    wf _ = 0

decode :: Signature -> Model -> Parameter
decode sig model =
  (decodeWeight model sig, decodePrecedence model)
