module Order where

import Data.List
import Term
import Rule
import TRS
import SMT
import Algebra
import qualified LPO
import qualified KBO
import qualified ELPO
import qualified EKBO
import qualified ELPOEKBO

-- Proving termination 

data Class =
    LPOClass
  | KBOClass
  | ELPOClass
  | EKBOClass
  | ELPOEKBOClass

instance Show Class where
  show LPOClass     = "LPO"
  show KBOClass      = "KBO"
  show ELPOClass     = "ELPO"
  show EKBOClass     = "EKBO"
  show ELPOEKBOClass = "ELPO+EKBO" 

data Instance =
    LPOInstance  Algebra.Precedence
  | KBOInstance  KBO.Parameter
  | ELPOInstance ELPO.Parameter 
  | EKBOInstance EKBO.Parameter
  | ELPOEKBOInstance ELPOEKBO.Parameter

instance Show Instance where
  show (LPOInstance parameter)      = LPO.showParameter parameter
  show (KBOInstance parameter)      = KBO.showParameter parameter
  show (ELPOInstance parameter)     = ELPO.showParameter parameter
  show (EKBOInstance parameter)     = EKBO.showParameter parameter
  show (ELPOEKBOInstance parameter) = show parameter

data Encoding = Orientable | Reducible | CPReducible deriving Show

classOf :: Instance -> Class
classOf (LPOInstance _)      = LPOClass
classOf (KBOInstance _)      = KBOClass
classOf (ELPOInstance _)     = ELPOClass
classOf (EKBOInstance _)     = EKBOClass
classOf (ELPOEKBOInstance (ELPOEKBO.ELPO _)) = ELPOClass
classOf (ELPOEKBOInstance (ELPOEKBO.EKBO _)) = EKBOClass

encode :: Class -> Signature -> TRS -> (Formula, [(Rule, Formula)])
encode LPOClass      sig trs = LPO.encode  sig trs
encode KBOClass      sig trs = KBO.encode  sig trs
encode ELPOClass     sig trs = ELPO.encode sig trs
encode EKBOClass     sig trs = EKBO.encode sig trs
encode ELPOEKBOClass sig trs = ELPOEKBO.encode sig trs

decode :: Class -> Signature -> Model -> Instance
decode LPOClass      sig model = LPOInstance      (LPO.decode     sig model)
decode KBOClass      sig model = KBOInstance      (KBO.decode     sig model)
decode ELPOClass     sig model = ELPOInstance     (ELPO.decode    sig model)
decode EKBOClass     sig model = EKBOInstance     (EKBO.decode    sig model)
decode ELPOEKBOClass sig model = ELPOEKBOInstance (ELPOEKBO.decode sig model)

gt :: Instance -> Term -> Term -> Bool
gt (LPOInstance param)      s t = LPO.gt      param s t
gt (KBOInstance param)      s t = KBO.gt      param s t
gt (ELPOInstance param)     s t = ELPO.gt     param s t
gt (EKBOInstance param)     s t = EKBO.gt     param s t
gt (ELPOEKBOInstance param) s t = ELPOEKBO.gt param s t

terminating :: String -> Class -> TRS -> IO (Maybe Instance)
terminating tool order trs = do
  smtOutput <- sat tool (conj (hard : [ f | (_, f) <- a ]))
  case smtOutput of
    Nothing    -> return Nothing
    Just model -> return (Just (decode order sig model)) 
  where
    (hard, a) = encode order (TRS.signatureOf trs) trs
    sig = TRS.signatureOf trs

-- Find k maximal subsystems

rewriteRules :: TRS -> TRS
rewriteRules es =
  nubBy Rule.variant
    [ rule | (s, t) <- es, rule <- [(s, t), (t, s)], Rule.variableCondition rule ]

excludeTRSs :: [(Rule, Formula)] -> [TRS] -> Formula
excludeTRSs a trss =
  conj [ disj [ f | (rule, f) <- a, not (any (Rule.variant rule) trs) ]
       | trs <- trss ]

reducibleTerm :: [(Rule, Formula)] -> Term -> Formula
reducibleTerm a t =
  disj [ f | (rule, f) <- a, reducible [rule] t ]

reducibleEq :: [(Rule, Formula)] -> Rule -> Formula
reducibleEq a (s, t)
  | s == t    = top
  | otherwise = disj [ f | (rule, f) <- a, reducible [rule] s || reducible [rule] t ]

orientableEq :: [(Rule, Formula)] -> Rule -> Formula
orientableEq a (s, t) =
  disj [ f | (rule, f) <- a, rule == (s, t) || rule == (t, s) ]

redundant :: [(Rule, Formula)] -> Rule -> Formula
redundant a (s, t)
  | s == t    = top
  | otherwise = 
      conj [ neg f 
           | rule <- [(s,t),(t,s)], 
             Just f <- [lookup rule a] ] 

orientableRule :: Rule -> Bool
orientableRule rule@(_, r) =
  Rule.variableCondition rule &&
  not (reducible [rule] r)

orientableRules :: TRS -> TRS
orientableRules cs = [ rule | rule <- rewriteRules cs, orientableRule rule ]

data Measure = Reducibility | Redundancy

constraints :: Class -> Measure -> TRS -> [TRS] -> ([(Rule,Formula)], Formula, [Formula])
constraints oc m cs trss =
  case m of
    Redundancy   -> (a, conj (orderConstraint : exclusion : reducibilities), [ redundant a e | e <- cs ])
    Reducibility -> (a, conj [orderConstraint, exclusion], reducibilities)
  where
    (orderConstraint, a) = encode oc (TRS.signatureOf cs) (orientableRules cs)
    exclusion = excludeTRSs a trss
    reducibilities = [ reducibleEq a e | e <- cs ]

newOrder :: String -> Class -> Measure -> TRS -> [TRS] -> IO (Maybe (TRS, Instance))
newOrder tool oc m cs trss = do
  smtOutput <- maxsat tool hard softs
  case smtOutput of
    Nothing -> return Nothing
    Just model -> 
      case [ rule | (rule, f) <- a, evalFormula model f ] of
        []  -> return Nothing
        trs -> return (Just (trs, decode oc (TRS.signatureOf cs) model))
  where
    (a, hard, softs) = constraints oc m cs trss


kOrders' :: String -> Class -> Measure -> Int -> TRS -> [(TRS, Instance)] -> IO [(TRS, Instance)]
kOrders' _    _  _ 0 _  pairs = return pairs
kOrders' tool oc m k cs pairs = do
  result <- newOrder tool oc m cs [ trs | (trs, _) <- pairs ]
  case result of
    Nothing        -> return pairs
    Just (trs, oi) -> kOrders' tool oc m (k - 1) cs (pairs ++ [(trs, oi)])

kOrders :: String -> Class -> Measure -> Int -> TRS -> IO [(TRS, Instance)]
kOrders tool oc m k cs = kOrders' tool oc m k cs []
